import pytest
import pytest_deprecate

from bobtail.wsgi import BobTail
from tests.fixtures import (
    bobtail_app,
    environ,
    multipart_data,
    form_data,
    multipart_data_with_file,
)

from bobtail.request import Request
from bobtail.headers import RequestHeaders
from bobtail.exceptions import FormDataError, MultipartFormDataError


class TestRequest:

    def test_get_path(self, bobtail_app, environ):
        class Images:
            def get(self, req, res):
                pass

        routes = [(Images(), "/images/{id:int}")]

        app = bobtail_app(routes=routes)

        env = environ(path="/images/1")
        data = app(env, lambda s, r: None)
        req_path = app.request.get_path()
        assert req_path == "/images/1"

        env = environ(path="/images?q=monalisa")
        data = app(env, lambda s, r: None)
        req_path = app.request.get_path()
        assert req_path == "/images?q=monalisa"

    def test_get_args(self, bobtail_app, environ):
        class Images:

            def get(self, req, res):
                pass

        routes = [
            (Images(), "/images/{id:int}/{name:str}/{is_raining:bool}")
        ]

        app = bobtail_app(routes=routes)

        env = environ(path="/images/1/hello/true")
        data = app(env, lambda s, r: None)
        id = app.request.get_arg("id")
        name = app.request.get_arg("name")
        is_raining = app.request.get_arg("is_raining")
        assert id == 1
        assert name == "hello"
        assert is_raining is True

    def test_get_args_if_key_not_present(self, bobtail_app, environ):
        class Images:

            def get(self, req, res):
                pass

        routes = [
            (Images(), "/images/{id:int}")
        ]

        app = bobtail_app(routes=routes)

        env = environ(path="/images/1")
        data = app(env, lambda s, r: None)
        assert app.request.get_arg("id") == 1
        assert app.request.get_arg("name") == None

        env = environ(path="/images")
        data = app(env, lambda s, r: None)
        assert app.request.get_arg("id") == None

    def test_get_json(self):
        req_headers = RequestHeaders("application/json")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=b'{\n    "name": "joe"\n, "email": "joe@email.com"\n}',
            headers=req_headers,
        )
        result = req.get_json()
        expected = {"email": "joe@email.com", "name": "joe"}
        assert result == expected

    def test_get_body(self):
        req_headers = RequestHeaders("application/json")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=b'test text',
            headers=req_headers,
        )
        result = req.get_body()
        expected = 'test text'
        assert result == expected

    def test_get_params(self):
        req_headers = RequestHeaders("application/json")
        req = Request(
            query_str="name=joe&age=48",
            path="/images",
            method="GET",
            byte_data=b'test text',
            headers=req_headers,
        )
        result = req.get_params()
        expected = {"name": "joe", "age": "48"}
        assert result == expected

    @pytest.mark.deprecated("This feature will be dropped in 0.1.0")
    def test_get_form_data(self, multipart_data):
        req_headers = RequestHeaders("application/x-www-form-urlencoded")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data,
            headers=req_headers,
        )
        result = req.get_form_data()
        expected = {
            "email": {
                "name": "email",
                "value": "test@test.com",
                "type": "text",
            },
            "password": {
                "name": "password",
                "value": "wizard",
                "type": "text",
            },
        }
        assert result == expected

    @pytest.mark.deprecated("This feature will be dropped in 0.1.0")
    def test_get_multipart_data(self, multipart_data):
        req_headers = RequestHeaders("multipart/form-data")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data,
            headers=req_headers,
        )
        result = req.get_multipart_data()
        expected = {
            "email": {
                "name": "email",
                "value": "test@test.com",
                "type": "text",
            },
            "password": {
                "name": "password",
                "value": "wizard",
                "type": "text",
            },
        }
        assert result == expected

    @pytest.mark.deprecated("This feature will be dropped in 0.1.0")
    def test_get_form_value(self, form_data):
        req_headers = RequestHeaders("application/x-www-form-urlencoded")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=form_data,
            headers=req_headers,
        )
        assert req.get_form_value("password") == "wizard"

        with pytest.raises(FormDataError):
            req.get_form_value("bananas")

    @pytest.mark.deprecated("This feature will be dropped in 0.1.0")
    def test_get_multipart_value(self, multipart_data):
        req_headers = RequestHeaders("multipart/form-data")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data,
            headers=req_headers,
        )
        assert req.get_multipart_value("email") == "test@test.com"

        with pytest.raises(MultipartFormDataError):
            req.get_multipart_value("bananas")

    @pytest.mark.deprecated("This feature will be dropped in 0.1.0")
    def test_get_filename_value(self, multipart_data_with_file):
        req_headers = RequestHeaders("multipart/form-data")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data_with_file,
            headers=req_headers,
        )
        assert req.get_filename_value("logo") == "bobtail.png"

        with pytest.raises(MultipartFormDataError):
            req.get_filename_value("bananas")



from .wsgi import BobTail
from .options import BaseOptions
from .route import AbstractRoute, Handler, Route
from .request import Request
from .response import Response
from .middleware import Middleware, AbstractMiddleware, Tail
from .exceptions import *



from bobtail.response import Response
from bobtail.options import BaseOptions
from tests.fixtures import default_options


class TestResponse:

    def test_headers(self, default_options):
        headers = {"Content-Type": "text/plain"}
        r = Response(default_options)
        r.set_headers(headers)
        assert r.headers == headers

    def test_status(self, default_options):
        status = 404
        r = Response(default_options)
        r.set_status(status)
        assert r.status == status

    def test_body(self, default_options):
        json_data = {"id": 1}
        r = Response(default_options)
        r.set_body(json_data)
        assert r.body == json_data

    def test_content_len(self, default_options):
        byte_data = b"hello, world!"
        r = Response(default_options)
        r._set_content_len(byte_data)
        expected = {
            "Content-Type": "application/json",
            "Content-Length": f"{len(byte_data)}",
        }
        assert expected == r.headers

    def test_set_html(self, default_options):
        r = Response(default_options)
        r.set_html("hello world")
        assert r.html == "hello world"
        assert r.headers == {'Content-Type': 'text/html'}

    def test_set_static(self, default_options):
        class _Option(BaseOptions):
            STATIC_DIR = "tests/static"
        r = Response(_Option())
        fp = "tests/static/img/cat1.jpg"
        r.set_static(fp)
        assert isinstance(r.static, bytes)



from typing import Dict, Union
from abc import ABC
from warnings import warn

from bobtail.wsgi_input import WSGIInput
from bobtail.headers import RequestHeaders
from bobtail.exceptions import FormDataError, MultipartFormDataError
from bobtail.form import AbstractForm, Form, MultipartForm


class Request(ABC):

    path: str

    method: str

    vars: Dict

    args: Dict

    headers: RequestHeaders

    byte_data: input

    wsgi_input: WSGIInput

    form: Form

    query_str: str

    multipart: MultipartForm

    def __init__(self, *,  # pylint: disable=too-many-arguments
                 path: str,
                 method: str,
                 byte_data: bytes,
                 headers: RequestHeaders,
                 query_str: str,
                 ):
        self.path = path
        self.method = method
        self.headers = headers
        self.args: Dict = {}
        self.query_str = query_str
        self.wsgi_input = WSGIInput(
            byte_data=byte_data,
            headers=self.headers,
        )
        self.form = Form(self.wsgi_input)
        self.multipart = MultipartForm(self.wsgi_input)

    def get_path(self) -> str:
        return self.path

    def get_arg(self, name: str) -> Union[str, int, bool]:
        """
        :param name:
        :type name:
        :return:
        :rtype:
        """
        if name not in self.args:
            return None

        arg_value = self.args[name]["value"]
        match self.args[name]["type"]:
            case "int":
                return int(arg_value)
            case "str":
                return arg_value
            case "bool":
                return bool(arg_value == "true" or arg_value == "True")

    def set_args(self, args):
        self.args = args

    def get_json(self) -> Dict:
        """
        :return:
        :rtype:
        """
        return self.wsgi_input.get_json()

    def get_body(self) -> str:
        """
        Handles text/plain
        :return:
        :rtype:
        """
        return self.wsgi_input.get_body()

    def get_form_data(self) -> Dict:
        """
        Handles application/x-www-form-urlencoded
        :return:
        :rtype:
        """
        warn(DeprecationWarning("[Bobtail]: Please use the Form API. See ..."))
        return self.wsgi_input.get_form_data()

    def get_multipart_data(self) -> Dict:
        """
        Handles multipart/form-data
        :return:
        :rtype:
        """
        warn(DeprecationWarning("[Bobtail]: Please use the MultipartForm API. See ..."))
        return self.wsgi_input.get_multipart_data()

    def get_form_value(self, name: str) -> str:
        """
        :param name:
        :return:
        """
        warn(DeprecationWarning("[Bobtail]: Please use the Form API. See ..."))
        try:
            data = self.get_form_data()
            return data[name]["value"]
        except KeyError as exc:
            raise FormDataError(
                f"Error getting form value for {name} field"
            ) from exc

    def get_multipart_value(self, name: str) -> str:
        """
        :param name:
        :return:
        """
        warn(DeprecationWarning("[Bobtail]: Please use the MultipartForm API. See ..."))
        try:
            data = self.get_multipart_data()
            return data[name]["value"]
        except KeyError as exc:
            raise MultipartFormDataError(
                f"Error getting form value for {name} field"
            ) from exc

    def get_filename_value(self, filename: str) -> str:
        """
        :param filename:
        :return:
        """
        warn(DeprecationWarning("[Bobtail]: Please use the MultipartForm API. See ..."))
        try:
            data = self.get_multipart_data()
            return data[filename]["value"]["filename"]
        except KeyError as exc:
            raise MultipartFormDataError(
                f"Filename Error: getting {filename} from multipart form data"
            ) from exc

    def get_params(self) -> Dict:
        """
        This method returns a dict og query params where the key
        is on the left side of the `=` sign & the value is pn the right.
        For example:

            # for route "/images?name=joe&age=48"

            def get(self, req: Request, res: Response):
                result = req.get_params() # {"name": "joe", "age": "48"}

        :return: Dict
        """
        param_dict = {}
        pl = self.query_str.split("&")
        for kv in pl:
            k, v = kv.split("=")
            param_dict[k] = v
        return param_dict



import pytest

from tests.fixtures import bobtail_app
from tests.fixtures import (
    route_class_one,
    route_class_two,
    middle_cors,
    middle_logger,
    environ,
)

from bobtail.wsgi import BobTail


class TestBobtail:

    app: BobTail = None

    def teardown_method(self):
        self.app = None

    def test_use(self, bobtail_app, middle_cors, middle_logger, environ):
        class Images:

            def get(self, req, res):
                res.set_headers({"Content-type": "text/plain"})
                res.set_status(200)
                res.set_body({})

        routes = [
            (Images(), "/images")
        ]

        self.app = bobtail_app(routes=routes)

        # Check no middleware is added
        assert self.app.middleware.middlewares is None

        self.app.use(middle_cors)
        self.app.use(middle_logger)

        assert len(self.app.middleware.middlewares) == 2

        env = environ()
        _ = self.app(env, lambda s, r: None)
        assert self.app.response.headers["Access-Control-Allow-Origin"] == "*"

    def test_handlers_get(self, bobtail_app, route_class_one, environ):

        routes = [
            (route_class_one, "/images")
        ]
        self.app = bobtail_app(routes=routes)

        # GET
        env = environ()
        data = self.app(env, lambda s, r: None)
        assert self.app.response.body == {"users": [{"id": 1}, {"id": 2}]}
        assert self.app.response.status == 200
        assert data == [b'{\n  "users": [\n    {\n      "id": 1\n    },\n    {\n      "id": 2\n    }\n' b'  ]\n}']
        # Check no middleware is added
        assert self.app.middleware.middlewares is None

    def test_handlers_post(self, bobtail_app, route_class_one, environ):
        routes = [
            (route_class_one, "/images")
        ]

        self.app = bobtail_app(routes=routes)

        # POST
        env = environ(method="POST")

        data = self.app(env, lambda s, r: None)

        assert data == []
        assert self.app.response.body is None
        assert self.app.response.status == 202
        # Check no middleware is added
        assert self.app.middleware.middlewares is None

    def test_handlers_delete(self, bobtail_app, route_class_one, environ):
        routes = [
            (route_class_one, "/images")
        ]
        self.app = bobtail_app(routes=routes)

        # # DELETE
        env = environ(method="DELETE")
        data = self.app(env, lambda s, r: None)
        assert data == []
        assert self.app.response.body is None
        assert self.app.response.status == 201
        # Check no middleware is added
        assert self.app.middleware.middlewares is None

    def test_handlers_put(self, bobtail_app, route_class_one, environ):
        routes = [
            (route_class_one, "/images")
        ]
        self.app = bobtail_app(routes=routes)

        # PUT
        env = environ(method="PUT")
        data = self.app(env, lambda s, r: None)
        assert data == [b'{\n  "id": 1\n}']
        assert self.app.response.body == {"id": 1}
        assert self.app.response.status == 202
        # Check no middleware is added
        assert self.app.middleware.middlewares is None

    def test_handlers_path(self, bobtail_app, route_class_one, environ):
        routes = [
            (route_class_one, "/images")
        ]
        self.app = bobtail_app(routes=routes)

        # PATCH
        env = environ(method="PATCH", query_str="name=joe&age=48")
        data = self.app(env, lambda s, r: None)
        assert data == [b'{\n  "id": 1\n}']
        assert self.app.response.body == {"id": 1}
        assert self.app.response.status == 202
        assert self.app.request.get_params() == {"name": "joe", "age": "48"}

        # Check no middleware is added
        assert self.app.middleware.middlewares is None

    def test_headers(self, bobtail_app, route_class_two, environ):

        routes = [
            (route_class_two, "/images")
        ]

        self.app = bobtail_app(routes=routes)

        # Check no middleware is added
        assert self.app.middleware.middlewares is None

        env = environ()
        _ = self.app(env, lambda s, r: None)
        assert self.app.response.headers["Content-type"] == "text/plain"

    def test_args_not_in_path(self, bobtail_app, route_class_two, environ):
         # Ref: https://github.com/joegasewicz/bobtail/issues/27
        routes = [
            (route_class_two, "/images/{id:int}")
        ]

        self.app = bobtail_app(routes=routes)

        # Check no middleware is added
        assert self.app.middleware.middlewares is None

        env = environ()
        data = self.app(env, lambda s, r: None)
        assert data is not None

    def test_no_options(self, environ):
        self.app = BobTail(routes=[])
        env = environ()
        try:
            _ = self.app(env, lambda s, r: None)
        except AttributeError:
            pytest.fail("Default options are not set!")

        assert self.app.options.PORT == 8000
        assert self.app.options.STATIC_DIR is "static"
        assert self.app.options.TEMPLATE_DIR is "templates"




from typing import List, Dict

from bobtail.response import Response
from bobtail.request import Request
from bobtail.status import Status
from bobtail.exceptions import NoRoutesError, RouteClassError
from bobtail.route import Route, Handler
from bobtail.parser import Parser
from bobtail.middleware import Middleware
from bobtail.headers import ResponseHeaders, RequestHeaders
from bobtail.options import BaseOptions
from bobtail.middleware import AbstractMiddleware


class BobTail:
    """
    :kwargs:
    :key routes: A tuple list of :class:`~AbstractRoute` & request path
    """
    #: Bobtail is a micro http web framework.
    environ: Dict

    #: Routes is a list of tuple :class:`~AbstractRoute` & request path pairs.
    #: For Example::
    #:
    #:  routes = [
    #:      (Images(), "/images")
    #:  ]
    #:
    routes: List[Route]

    #: Options (Optional). See :class:`~BaseOptions` for option list. Base options
    #: can be overridden & or set on a concrete BaseOptions instance.
    #: For Example::
    #:
    #:  from bobtail.options import BaseOptions
    #:
    #:  class Options(BaseOptions):
    #:      PORT = 8000                     # set by default
    #:      STATIC_DIR = "app/static"       # set by default
    #:      TEMPLATE_DIR = "app/templates"  # set by default
    #:
    #:  app = Bobtail(Options, routes=[])
    #:
    options: BaseOptions

    response: Response = None

    request: Request

    _status: Status

    _body: str

    parse_metadata: Dict = None

    middleware: Middleware = None

    request_headers: RequestHeaders

    response_headers: ResponseHeaders

    def __init__(self, *args, **kwargs):
        if "routes" not in kwargs:
            raise NoRoutesError("Expected a list of routes")
        self.routes = kwargs["routes"]
        _options = kwargs.get("options")
        if _options:
            self.options = _options
        else:
            class DefaultOptions(BaseOptions):
                pass

            self.options = DefaultOptions()
        self.middleware = Middleware()

    def _handle_404(self, req: Request, res: Response):
        self.response.set_status(404)

    def _init_response(self):
        self.response = Response(self.options)

    def _set_request(self):
        self.request = Request(
            path=self.environ["PATH_INFO"],
            method=self.environ["REQUEST_METHOD"],
            byte_data=self.environ["wsgi.input"].read(),
            headers=RequestHeaders(content_type=self.environ.get("CONTENT_TYPE")),
            query_str=self.environ["QUERY_STRING"]
        )

    def _call_handler(self, route: callable, method: str):
        if hasattr(route,  method):
            handler: Handler = getattr(route, method)
            if not handler:
                self.middleware.call(self.request, self.response, self._handle_404)
                return
            self.middleware.call(self.request, self.response, handler)
        else:
            self.middleware.call(self.request, self.response, self._handle_404)

    def _handle_route(self):
        p = Parser(self.routes, self.request.path)
        self.parse_metadata = p.route()
        # Set the args on the request object
        if self.parse_metadata and "vars" in self.parse_metadata:
            self.request.set_args(self.parse_metadata["vars"])
        for current_route in self.routes:
            route, _ = current_route
            if route.__class__.__name__ == p.get_matched():
                match self.request.method:
                    case "GET":
                        self._call_handler(route, "get")
                        return
                    case "POST":
                        self._call_handler(route, "post")
                        return
                    case "DELETE":
                        self._call_handler(route, "delete")
                        return
                    case "PUT":
                        self._call_handler(route, "put")
                        return
                    case "PATCH":
                        self._call_handler(route, "patch")
                        return
        self.middleware.call(self.request, self.response, self._handle_404)

    def __call__(self, environ, start_response):
        self.environ = environ

        # Set request & response
        self._set_request()
        self._init_response()
        # Call route handler with default response
        self._handle_route()

        self._status = Status()
        status = self._status.get(self.response.status)

        response_headers = []

        for k, v in self.response.headers.items():
            t = (k, v,)
            response_headers.insert(0, t)

        # Start response
        start_response(status, response_headers)
        # Process the final byte list & headers
        data = self.response._process()
        # clean up text/html string
        return data

    def use(self, middleware: AbstractMiddleware):
        """
        Enables using third party middleware.
        For example::

            from bobttail_logger import BobtailLogger

            app = Bobtail(routes=routes)

            # Here we are using `bobtail-logger` logging middleware
            app.use(BobtailLogger())

        Creating custom middleware example.
        A Middleware object must implement :class:`AbstractMiddleware`.
        For example::

            from bobtail import Request, Response
            from bobtail.middleware import AbstractMiddleware, Tail

            class BobtailCors(AbstractMiddleware):

                def run(self, req: Request, res: Response, tail: Tail) -> None:
                    res.set_headers({
                        "Access-Control-Allow-Origin": "*",
                    })
                    tail(req, res)

        :param middleware: :class:`AbstractMiddleware`
        :return: None
        """
        self.middleware.add(middleware)



import pytest

from bobtail.parser import Parser
from bobtail.route import AbstractRoute


class TestParser:

    def test_route(self):
        class Images1:
            def get(self, req, res):
                pass

        class Videos1:
            def get(self, req, res):
                pass

        class Files1:
            def get(self, req, res):
                pass

        routes1 = [
            (Images1(), "/images/image/{id:int}"),
            (Videos1(), "/videos/movie/{name:string}"),
            (Files1(), "/files/new/{true:bool}"),
        ]
        path1 = "/images/image/1"
        p1 = Parser(routes1, path1)
        result = p1.route()
        expected = {
            'route': '/images/image/{id:int}',
            'split': ['images', 'image', '{id:int}'],
            'vars': {
                'id': {
                    'name': 'id',
                    'type': 'int',
                    'value': '1',
                }
            },
            'path': '/images/image/1',
            'class': 'Images1'
        }
        assert result == expected

    def test_route_with_multiple_vars(self):
        class Images2:
            def get(self, req, res):
                pass

        class Videos2:
            def get(self, req, res):
                pass

        class Files2:
            def get(self, req, res):
                pass

        routes2 = [
            (Images2(), "/images/image/{id:int}/pic/{file_name:str}"),
            (Videos2(), "/videos/movie/{name:str}"),
            (Files2(), "/files/new/{true:bool}"),
        ]
        path2 = "/images/image/1/pic/sunny.png"
        p2 = Parser(routes2, path2)
        result2 = p2.route()
        expected2 = {
            'route': '/images/image/{id:int}/pic/{file_name:str}',
            'split': ['images', 'image', '{id:int}', 'pic', '{file_name:str}'],
            'vars': {
                'id': {
                    'name': 'id',
                    'type': 'int',
                    'value': '1',
                },
                'file_name': {
                    'name': 'file_name',
                    'type': 'str',
                    'value': 'sunny.png',
                }
            },
            'path': '/images/image/1/pic/sunny.png',
            'class': 'Images2'
        }
        assert result2 == expected2

    def test_route_index_route(self):
        class Home3:
            def get(self, req, res):
                pass

        class Images3:
            def get(self, req, res):
                pass

        class Files3:
            def get(self, req, res):
                pass

        routes3 = [
            (Images3(), "/images"),
            (Home3(), "/"),
            (Files3(), "/"),
        ]
        path = "/"
        p3 = Parser(routes3, path)
        result = p3.route()
        expected = {
            'class': 'Home3',
            'path': '/',
            'route': '/',
            'split': [''],
            'vars': None,
        }
        assert result == expected

    def test_extra_segments(self):
        class Images6:
            def get(self, req, res):
                pass

        class Videos6:
            def get(self, req, res):
                pass

        class Files6:
            def get(self, req, res):
                pass

        routes2 = [
            (Images6(), "/images/image/{id:int}/pic/{file_name:str}"),
            (Videos6(), "/videos/movie/{name:str}"),
            (Files6(), "/files/new/{true:bool}"),
        ]
        path2 = "/images/image/1/pic/sunny.png/hello!"
        p2 = Parser(routes2, path2)
        _ = p2.route()
        assert p2.meta_data["matched"] is None

    def test_get_matched(self):
        class Home4:
            def get(self, req, res):
                pass

        class Images4:
            def get(self, req, res):
                pass

        class Files4:
            def get(self, req, res):
                pass

        routes4 = [
            (Images4(), "/images"),
            (Home4(), "/"),
            (Files4(), "/"),
        ]
        path = "/"
        p4 = Parser(routes4, path)
        result = p4.route()
        expected = {
            'class': 'Home4',
            'path': '/',
            'route': '/',
            'split': [''],
            'vars': None,
        }
        assert result == expected
        m = p4.get_matched()
        assert m == "Home4"


    def test_static_route(self):
        class Images6:
            def get(self, req, res):
                pass

        class Videos6:
            def get(self, req, res):
                pass

        class Files6:
            def get(self, req, res):
                pass

        routes2 = [
            (Videos6(), "/videos"),
            (Files6(), "/files/new/{true:bool}"),
            (Images6(), "/static/*"),
        ]
        path2 = "/static/images/sunny.png"
        p2 = Parser(routes2, path2)
        _ = p2.route()
        assert p2.meta_data["matched"] == "Images6"



class InvalidStatusError(Exception):
    pass


class NoRoutesError(Exception):
    pass


class RouteClassError(Exception):
    pass


class FormDataError(Exception):
    pass


class MultipartFormDataError(Exception):
    pass


class TemplatePathError(Exception):
    pass


class StaticFileError(Exception):
    pass



from typing import Dict, List, Optional
import json

from bobtail.options import BaseOptions
from bobtail.exceptions import StaticFileError
from bobtail.static_files import (
    AUDIO_FILETYPE ,
    IMAGE_FILETYPE,
    StaticFiles,
    TXT_FILETYPE,
    VIDEO_FILETYPE,
)

class Response:
    """
    The :class:`Response` object is available from within a Route method.
    """
    status: int = 200

    body: Dict = None

    html: str = None

    static: bytes = None

    options: BaseOptions

    headers: Dict = {
        "Content-Type": "application/json",
    }

    _content_length: int = None

    def __init__(self, options: BaseOptions):
        self.options = options

    def _process(self) -> List[bytes]:
        """Process final response"""
        if self.body is not None:
            resp_data = bytes(json.dumps(self.body, indent=2), "utf-8")
            self._set_content_len(resp_data)
            return [resp_data]
        if self.html is not None:
            resp_data = bytes(self.html, "utf-8")
            self._set_content_len(resp_data)
            return [resp_data]
        if self.static is not None:
            self._set_content_len(self.static)
            return [self.static]
        self._set_content_len(None)
        return []

    def _set_content_len(self, data: Optional[bytes]) -> None:
        if self._content_length is None:
            con_len = 0 if data is None else len(data)
            self.headers = {
                **self.headers,
                "Content-Length": f"{con_len}",
            }

    def set_headers(self, value: Dict) -> None:
        """
        Headers can be set from within a route method. The `Content-Type` header get
        set dynamically from :class:`Response` methods. By default, `Content-Type` is
        set to `application/json`.
        For Example::

            def get(self, req: Request, res: Response):
                res.set_headers({
                    "Content-Type": "application/json",
                })

        :param value: Excepts a dict
        """
        self.headers = {
            **self.headers,
            **value,
        }

    def set_status(self, value: int) -> None:
        """
        Set the response status.
        You can set the status with the Response object's set_status method.
        The default status is always set to 200 if there are no errors.
        For Example::

            def get(self, req: Request, res: Response):
                res.set_headers({"Content-type": "text/plain"})

        :param value: Sets the response status
        """
        self.status = value

    def set_body(self, value: Dict) -> None:
        """
        Set the body of the request.
        For example::

            def get(self, req: Request, res: Response):
                res.set_body({id: 1})

        :param value:
        """
        self.body = value

    def set_html(self, template_str: str) -> None:
        """
        Renders an HTML string. See :class:`~BaseOptions.TEMPLATE_DIR` for the
        template directory path.
        For example::

            template_str = "<h1>Hello!</h1>"

            def get(self, req: Request, res: Response):
                res.set_html(template_str)

        :param template_str:
        """
        self.set_headers({"Content-Type": "text/html"})
        self.html = template_str

    def set_static(self, path: str):
        """
        To declare a static route postfix a `*` to the route's path::

            from bobtail import BobTail AbstractRoute, BaseOptions
            from bobtail_jinja2 import BobtailJinja2

            routes = [
                (Static(), "/static/*"),
            ]

            class Options(BaseOptions):
                STATIC_DIR = "app/static"
                TEMPLATE_DIR = "app/templates"

            blog = BobTail(routes=routes, options=Options())
            blog.use(BobtailJinja2(template_dir="app/templates"))

        Calling `set_static` from within a route method will render a static
        file such as a .css, .js or a media type file. The :class:`~BaseOptions`
        class sets the `STATIC_DIR` directory.
        For example::

                class Static(AbstractRoute):
                    def get(self, req: Request, res: Response) -> None:
                        res.set_static(req.path)

        You can set the static file path using the :class:`~BaseOptions`.
        For example::

                class Options(BaseOptions):
                    STATIC_DIR = "/static"

                # Now in a route handler we can access static directory the via options
                class Static(AbstractRoute):
                    def get(self, req: Request, res: Response) -> None:
                        res.set_static(req.path)

        By default, `STATIC_DIR` is set to `/static`, if your static file is nested
        within a Python package, for example `app/static` the set as `STATIC_DIR = "app/static"`

        To render an image from within a Jinja2 template include the full path including the
        static directory name or path. For example::

            <!-- if STATIC_DIR = "/static" -->
            <body>
                <img src="/static/imgs/cat1.jpg" />
            </body>

        OR without the first forward slash::

            <body>
                <img src="static/imgs/cat1.jpg" />
            </body>

        :param path:
        """
        if len(path.split(".")) < 1:
            return None
        path_seg = path.split("/static")
        if len(path_seg) <= 1:
            return None
        path = path_seg[1]
        try:
            file_suffix = path.split("/")[-1:][0].split(".")[-1:][0]
            path = f"{self.options.STATIC_DIR}{path}"

            if file_suffix in IMAGE_FILETYPE:
                StaticFiles(self, "image").set_headers(file_suffix)
            elif file_suffix == "css":
                StaticFiles(self, "text").set_headers(file_suffix)
            elif file_suffix in VIDEO_FILETYPE:
                StaticFiles(self, "video").set_headers(file_suffix)
            elif file_suffix in AUDIO_FILETYPE:
                StaticFiles(self, "audio").set_headers(file_suffix)
            else:
                StaticFiles(self, "text").set_headers("plain")

        except Exception as exc:
            self.set_status(500)
            raise StaticFileError(
                f"Error getting filetype from static file path - {path}"
            ) from exc
        try:
            with open(path, "rb") as f:
                file_data = f.read()
                f.close()
            self.static = file_data
        except FileNotFoundError:
            self.set_status(404)



from abc import ABC


class BaseOptions(ABC):
    """
    To define port, static directory, template directory etc. you can
    create a concrete version of the BaseOptions abstract class.
    For Example::

        from bobtail.options import BaseOptions

        class Options(BaseOptions):
            PORT = 8001

        app = Bobtail(Options)

    """
    #: The port by default is set to `8000`
    PORT = 8000

    #: The static directory relative path is set by default to `static`
    STATIC_DIR = "static"

    #: The template directory relative path is set by default to `templates`
    TEMPLATE_DIR = "templates"



from abc import ABC, abstractmethod
from typing import Dict

from bobtail.wsgi_input import WSGIInput
from bobtail.exceptions import MultipartFormDataError, FormDataError


class AbstractForm(ABC):

    wsgi_input: WSGIInput

    def __init__(self, wsgi_input: WSGIInput):
        self.wsgi_input = wsgi_input

    @abstractmethod
    def get_field(self,field_name: str) -> str:
        pass


class Form(AbstractForm):

    def get_field(self,field_name: str) -> str:
        """
        :param field_name:
        :return str:
        """
        try:
            data = self.wsgi_input.get_form_data()
            return data[field_name]["value"]
        except KeyError as exc:
            raise FormDataError(
                f"Error getting form value for {field_name} field"
            ) from exc


class MultipartForm(AbstractForm):

    def get_field(self, field_name) -> str:
        """
        :param field_name:
        :return:
        """
        try:
            data = self.wsgi_input.get_multipart_data()
            return data[field_name]["value"]
        except KeyError as exc:
            raise MultipartFormDataError(
                f"Error getting form value for {field_name} field"
            ) from exc

    def get_file(self,field_name: str) -> Dict:
        """
        :param field_name:
        :return:
        """
        try:
            data = self.wsgi_input.get_multipart_data()
            return {
                "filename": data[field_name]["value"]["filename"],
                "data": data[field_name]["value"]["file_data"],
                "mimetype": data[field_name]["value"]["mimetype"],
            }
        except KeyError as exc:
            raise MultipartFormDataError(
                f"Error getting form value for {field_name} file"
            ) from exc

    def get_name(self,field_name: str) -> str:
        """
        :param field_name:
        :return:
        """
        try:
            data = self.wsgi_input.get_multipart_data()
            return data[field_name]["value"]["filename"]
        except KeyError as exc:
            raise MultipartFormDataError(
                f"Error getting filename for {field_name} file"
            ) from exc

    def get_data(self,field_name: str) -> bytes:
        """
        :param field_name:
        :return:
        """
        try:
            data = self.wsgi_input.get_multipart_data()
            return data[field_name]["value"]["file_data"]
        except KeyError as exc:
            raise MultipartFormDataError(
                f"Error getting file data for {field_name} file"
            ) from exc

    def get_mimetype(self,field_name: str) -> str:
        """
        :param field_name:
        :return:
        """
        try:
            data = self.wsgi_input.get_multipart_data()
            return data[field_name]["value"]["mimetype"]
        except KeyError as exc:
            raise MultipartFormDataError(
                f"Error getting file mimetype for {field_name} file"
            ) from exc



import pytest

from bobtail.wsgi import BobTail
from tests.fixtures import (
    bobtail_app,
    environ,
    multipart_data,
    form_data,
    multipart_data_with_file,
)
from bobtail.request import Request
from bobtail.headers import RequestHeaders
from bobtail.exceptions import FormDataError, MultipartFormDataError


class TestForm:

    def test_get_field(self, form_data):
        req_headers = RequestHeaders("application/x-www-form-urlencoded")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=form_data,
            headers=req_headers,
        )
        assert req.form.get_field("password") == "wizard"

        with pytest.raises(FormDataError):
            req.form.get_field("bananas")

class TestMultipartForm:

    def test_get_field(self, multipart_data):
        req_headers = RequestHeaders("multipart/form-data")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data,
            headers=req_headers,
        )
        assert req.multipart.get_field("email") == "test@test.com"

        with pytest.raises(MultipartFormDataError):
            req.multipart.get_field("bananas")

    def test_get_file(self, multipart_data_with_file):
        req_headers = RequestHeaders("multipart/form-data")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data_with_file,
            headers=req_headers,
        )
        result = req.multipart.get_file("logo")
        assert result["filename"] == "bobtail.png"
        assert result["mimetype"] == "image/png"
        assert isinstance(result["data"], bytes)


    def test_get_name(self, multipart_data_with_file):
        req_headers = RequestHeaders("multipart/form-data")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data_with_file,
            headers=req_headers,
        )
        assert req.multipart.get_name("logo") == "bobtail.png"

        with pytest.raises(MultipartFormDataError):
            req.multipart.get_name("bananas")

    def test_get_data(self, multipart_data_with_file):
        req_headers = RequestHeaders("multipart/form-data")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data_with_file,
            headers=req_headers,
        )
        result = req.multipart.get_data("logo")
        assert isinstance(result, bytes)

        with pytest.raises(MultipartFormDataError):
            req.multipart.get_data("bananas")

    def test_get_mimetype(self, multipart_data_with_file):
        req_headers = RequestHeaders("multipart/form-data")
        req = Request(
            query_str="",
            path="/images",
            method="POST",
            byte_data=multipart_data_with_file,
            headers=req_headers,
        )
        result = req.multipart.get_mimetype("logo")
        assert result == "image/png"

        with pytest.raises(MultipartFormDataError):
            req.multipart.get_mimetype("bananas")



from typing import Any


IMAGE_FILETYPE = (
    "jpg",
    "jpeg",
    "png",
    "gif",
    "tiff",
    "svg+xml",
    "x-icon",
)

TXT_FILETYPE = (
    "css",
    "csv",
    "html",
    "javascript",
    "plain",
    "xml",
)

VIDEO_FILETYPE = (
    "mpeg",
    "mp4",
    "quicktime",
    "webm",
    "x-ms-wmv",
    "x-msvideo",
    "x-flv",
)

AUDIO_FILETYPE = (
    "mpeg",
    "x-ms-wma",
    "vnd.rn-realaudio",
    "x-wav",
)


class StaticFiles:
    response: Any  # Response
    mimetypes: str

    def __init__(self, response: Any, mimetype: str):
        self.response = response
        self.mimetype = mimetype

    def set_headers(self, file_type: str) -> None:
        self.response.set_headers({"Content-Type": f"{self.mimetype}/{file_type}"})



import pytest

from bobtail.exceptions import InvalidStatusError
from bobtail.status import Status


class TestStatus:

    expected_code_with_message = {
        100: "Continue",
        101: "Switching Protocols",
        200: "OK",
        201: "Created",
        202: "Accepted",
        203: "Non-Authoritative Information",
        204: "No Content",
        205: "Reset Content",
        206: "Partial Content",
        300: "Multiple Choices",
        301: "Moved Permanently",
        302: "Found",
        303: "See Other",
        304: "Not Modified",
        305: "Use Proxy",
        307: "Temporary Redirect",
        400: "Bad Request",
        401: "Unauthorized",
        402: "Payment Required",
        403: "Forbidden",
        404: "Not Found",
        405: "Method Not Allowed",
        406: "Not Acceptable",
        407: "Proxy Authentication Required",
        408: "Request Timeout",
        409: "Conflict",
        410: "Gone",
        411: "Length Required",
        412: "Precondition Failed",
        413: "Request Entity Too Large",
        414: "Request-URI Too Long",
        415: "Unsupported Media Type",
        416: "Requested Range Not Satisfiable",
        417: "Expectation Failed",
        500: "Internal Server Error",
        501: "Not Implemented",
        502: "Bad Gateway",
        503: "Service Unavailable",
        504: "Gateway Timeout",
        505: "HTTP Version Not Supported",
    }

    def test_get(self):
        s = Status()
        for code, message in self.expected_code_with_message.items():
            expected_response = f"{str(code)} {message}"
            actual_response = s.get(code)
            assert expected_response == actual_response

    def test_get_on_invalid_status_code(self):
        with pytest.raises(InvalidStatusError):
            Status().get(0)
